"""
Module to parse xmind file into test suite and test case objects.
"""
import logging
import re
from xml.etree import ElementTree as  ET
from xml.etree.ElementTree import Element
from zipfile import ZipFile

from .datatype import *

content_xml = "content.xml"
comments_xml = "comments.xml"


def parse_xmind_file(file_path):
    """Extract xmind as zip file then read the content.xml"""
    with ZipFile(file_path) as xmind:
        for f in xmind.namelist():
            if f == content_xml:
                cache[content_xml] = xmind.open(f).read().decode()

            if f == comments_xml:
                cache[comments_xml] = xmind.open(f).read().decode()

    return parse_xmind_content()


def parse_xmind_content():
    """Main function to read the content xml and return test suite data."""
    xml_root = xmind_content_to_etree(cache[content_xml])
    assert isinstance(xml_root, Element)

    try:
        xml_root_suite = xml_root.find('sheet').find('topic')
        logging.info("Parse topic: {}".format(title_of(xml_root_suite)))
    except:
        logging.error('Cannot find any topic in your xmind!')
        raise

    root_suite = TestSuite()
    root_suite.sub_suites = []
    suite_nodes = children_topics_of(xml_root_suite)

    if not suite_nodes:
        raise ValueError("Cannot find any test suite in xmind!")

    for node in suite_nodes:
        suite = parse_suite(node)
        root_suite.sub_suites.append(suite)

    return root_suite


def xmind_content_to_etree(content):
    # Remove the default namespace definition (xmlns="http://some/namespace")
    xml_content = re.sub(r'\sxmlns="[^"]+"', '', content, count=1)
    return ET.fromstring(xml_content)


def xmind_xml_to_etree(xml_path):
    with open(xml_path) as f:
        content = f.read()
        return xmind_content_to_etree(content)


def comments_of(node):
    if cache.get(comments_xml, None):
        xml_root = xmind_content_to_etree(cache[comments_xml])
        node_id = node.attrib['id']
        comment = xml_root.find('./comment[@object-id="{}"]'.format(node_id))

        if comment:
            return comment.find('content').text


def title_of(node):
    return node.find('title').text


def note_of(topic_node):
    note_node = topic_node.find('notes')

    if note_node:
        note = note_node.find('plain').text
        return note.strip()


def maker_of(topic_node, maker_prefix):
    maker_node = topic_node.find('marker-refs')
    if maker_node:
        for maker in maker_node:
            maker_id = maker.attrib['marker-id']
            if maker_id.startswith(maker_prefix):
                return maker_id


def children_topics_of(topic_node):
    children = topic_node.find('children')

    if children:
        return children.find('./topics[@type="attached"]')


def parse_step(step_node):
    step = TestStep()
    step.action = title_of(step_node)
    expected_node = children_topics_of(step_node)

    if expected_node:
        step.expected = title_of(children_topics_of(step_node)[0])

    return step


def parse_steps(steps_node):
    steps = []

    for step_number, step_node in enumerate(steps_node, 1):
        step = parse_step(step_node)
        step.number = step_number
        steps.append(step)

    return steps


def parse_testcase(testcase_node):
    testcase = TestCase()
    testcase.name = title_of(testcase_node)
    testcase.summary = note_of(testcase_node)
    testcase.importance = maker_of(testcase_node, 'priority')
    testcase.preconditions = comments_of(testcase_node)
    steps_node = children_topics_of(testcase_node)

    if steps_node:
        testcase.steps = parse_steps(steps_node)

    return testcase


def parse_suite(suite_node):
    suite = TestSuite()
    suite.name = title_of(suite_node)
    suite.details = note_of(suite_node)
    suite.testcase_list = []
    testcase_nodes = children_topics_of(suite_node)

    if testcase_nodes:
        for node in testcase_nodes:
            testcase = parse_testcase(node)
            suite.testcase_list.append(testcase)

    return suite



"""
Module to parse test suite objects into testlink xml.
"""
import os
from io import BytesIO
from os.path import exists
from xml.dom import minidom
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement

from .datatype import *


class Tags():
    xml = 'xml'
    testsuite = "testsuite"
    details = 'details'
    testcase = 'testcase'
    summary = 'summary'
    precoditions = 'preconditions'
    steps = 'steps'
    step = 'step'
    step_number = 'step_number'
    actions = 'actions'
    expected = 'expectedresults'
    execution_type = 'execution_type'
    importance = 'importance'


class Attributes():
    name = 'name'


def to_testlink_xml_file(testsuite, path_to_xml):
    """Save test suite object to testlink xml file."""
    content = to_testlink_xml_content(testsuite)
    if exists(path_to_xml):
        os.remove(path_to_xml)

    with open(path_to_xml, 'w') as f:
        f.write(prettify_xml(content))


def _convert_importance(importance_value):
    mapping = {'priority-1': '3', 'priority-2': '2', 'priority-3': '1'}
    if importance_value in mapping.keys():
        return mapping[importance_value]
    else:
        return '2'


def to_testlink_xml_content(testsuite):
    assert isinstance(testsuite, TestSuite)
    root_suite = Element(Tags.testsuite)
    root_suite.set(Attributes.name, testsuite.name)
    cache['testcase_count'] = 0

    for suite in testsuite.sub_suites:
        assert isinstance(suite, TestSuite)

        if suite.name.startswith('!'):
            continue

        suite_element = SubElement(root_suite, Tags.testsuite)
        suite_element.set(Attributes.name, suite.name)

        if suite.details and not suite.details.startswith('!'):
            e = SubElement(suite_element, Tags.details)
            e.text = suite.details

        for testcase in suite.testcase_list:
            assert isinstance(testcase, TestCase)

            if testcase.name.startswith('!'):
                continue

            cache['testcase_count'] += 1
            testcase_element = SubElement(suite_element, Tags.testcase)
            testcase_element.set(Attributes.name, testcase.name)

            if testcase.summary and not testcase.summary.startswith('!'):
                e = SubElement(testcase_element, Tags.summary)
                e.text = testcase.summary

            if testcase.preconditions and not testcase.preconditions.startswith('!'):
                e = SubElement(testcase_element, Tags.precoditions)
                e.text = testcase.preconditions

            if testcase.execution_type:
                e = SubElement(testcase_element, Tags.execution_type)
                e.text = testcase.execution_type

            e = SubElement(testcase_element, Tags.importance)
            e.text = _convert_importance(testcase.importance)

            if testcase.steps:
                steps_element = SubElement(testcase_element, Tags.steps)

                for step in testcase.steps:
                    assert isinstance(step, TestStep)

                    if step.action.startswith('!'):
                        continue
                    else:
                        step_element = SubElement(steps_element, Tags.step)

                    if step.action and not step.action.startswith('!'):
                        e = SubElement(step_element, Tags.actions)
                        e.text = step.action

                    if step.expected and not step.expected.startswith('!'):
                        e = SubElement(step_element, Tags.expected)
                        e.text = step.expected

                    if step.execution_type:
                        e = SubElement(step_element, Tags.execution_type)
                        e.text = step.execution_type

                    e = SubElement(step_element, Tags.step_number)
                    e.text = str(step.number)

    tree = ElementTree.ElementTree(root_suite)
    f = BytesIO()
    tree.write(f, encoding='utf-8', xml_declaration=True)
    return f.getvalue()


def prettify_xml(xml_string):
    """Return a pretty-printed XML string for the Element.
    """
    reparsed = minidom.parseString(xml_string)
    return reparsed.toprettyxml(indent="\t")



"""
A tool to parse xmind file into testlink xml file, which will help
you generate a testlink recognized xml file, then you can import it
into testlink as test suites.

Usage:
 xmind2testlink [path_to_xmind_file]

Example:
 xmind2testlink C:\\tests\\testcase.xmind

"""

import sys

from .datatype import cache
from .testlink_parser import to_testlink_xml_file, to_testlink_xml_content
from .xmind_parser import parse_xmind_file


def xmind_to_testlink(xmind):
    xml_out = xmind[:-5] + 'xml'
    suite = parse_xmind_file(xmind)
    to_testlink_xml_file(suite, xml_out)
    return xml_out


def get_testcase_count(xmind):
    suite = parse_xmind_file(xmind)
    to_testlink_xml_content(suite)
    return cache.get('testcase_count', 0)


def main():
    if len(sys.argv) > 1 and sys.argv[1].endswith('.xmind'):
        xmind = sys.argv[1]
        xml_out = xmind_to_testlink(xmind)
        print("Generated: {}".format(xml_out))
    else:
        print(__doc__)


if __name__ == '__main__':
    main()



"""
Documentation
-------------
xmind2testlink is a tool to help you convert xmind file to testlink recognized xml files,
then you can import it into testlink as test suite and test cases. For more detail, please
go to: https://github.com/tobyqin/xmind2testlink

"""
from codecs import open
from os import path

from setuptools import setup, find_packages

current_dir = path.abspath(path.dirname(__file__))
long_description = __doc__

with open(path.join(current_dir, "CHANGELOG.md"), encoding="utf-8") as f:
    long_description += "\n" + f.read()

classifiers = ["License :: OSI Approved :: MIT License",
               "Topic :: Software Development",
               "Topic :: Utilities",
               "Operating System :: Microsoft :: Windows",
               "Operating System :: MacOS :: MacOS X"] + [
                  ("Programming Language :: Python :: %s" % x) for x in "2.7 3.5".split()]


def make_cmdline_entry_points():
    target = "src.xmind2testlink:main"
    entry_points = []
    entry_points.append("xmind2testlink=%s" % target)
    return entry_points


def main():
    setup(
        name="xmind2testlink",
        description="Convert xmind to TestLink xml",
        keywords="xmind testlink import converter testing testcase",
        long_description=long_description,
        classifiers=classifiers,
        version="1.0.5",
        author="Toby Qin",
        author_email="toby.qin@live.com",
        url="https://github.com/tobyqin/xmind2testlink",
        packages=find_packages(exclude=['tests', 'tests.*']),
        package_data={},
        install_requires=[],
        entry_points={
            "console_scripts": make_cmdline_entry_points(),
        },
        zip_safe=False,
    )


if __name__ == "__main__":
    main()



class TestSuite():
    sub_suites = None
    name = ""
    details = ""
    testcase_list = None


class TestCase():
    name = ""
    summary = ""
    preconditions = ""
    importance = 2
    execution_type = ""
    steps = None


class TestStep():
    number = 1
    action = ""
    expected = ""
    execution_type = ""


cache = {}
